@@ -2,35 +2,38 @@
 
 #' Construct conditional survival estimator Q_n
 #'
+#' @param type One of c("true", "Cox", "Random Forest", "Super Learner")
 #' @param dat Subsample of dataset returned by `ss` for which z==1
 #' @param vals List of values to pre-compute function on; REQUIRED FOR SUPERLEARNER
-#' @param type One of c("true", "Cox", "Random Forest", "Super Learner")
 #' @param return_model Logical; if TRUE, return the model object instead of the
 #'     function
-#' @param print_coeffs Logical; if TRUE, print the algorithm coefficients for
-#'     the conditional survival and censoring functions (applicable only if
-#'     type=="Super Learner")
 #' @return Conditional density estimator function
 #'
 #' @noRd
-construct_Q_n <- function(type, dat, vals, return_model=F, print_coeffs=F) {
+construct_Q_n <- function(type, dat_v, vals, return_model=F) {
+
+  dim_x <- attr(dat_v,"dim_x")
 
   if (type=="Cox") {
 
+    if (F) { weights <- NA } # Prevents CRAN error
+    x_names <- names(dat_v)[1:dim_x]
+    dat_v$delta2 <- 1 - dat_v$delta
+
     model_srv <- survival::coxph(
       formula = stats::formula(paste0("survival::Surv(y,delta)~",
-                               paste(names(dat$x),collapse="+"),"+s")),
-      data = cbind(y=dat$y, delta=dat$delta, dat$x, s=dat$s),
-      weights = dat$weights
+                               paste(x_names,collapse="+"),"+s")),
+      data = dat_v,
+      weights = weights
     )
     coeffs_srv <- as.numeric(model_srv$coefficients)
     bh_srv <- survival::basehaz(model_srv, centered=F)
 
     model_cens <- survival::coxph(
-      formula = stats::formula(paste0("survival::Surv(y,delta)~",
-                               paste(names(dat$x),collapse="+"),"+s")),
-      data = cbind(y=dat$y, delta=1-dat$delta, dat$x, s=dat$s),
-      weights = dat$weights
+      formula = stats::formula(paste0("survival::Surv(y,delta2)~",
+                               paste(x_names,collapse="+"),"+s")),
+      data = dat_v,
+      weights = weights
     )
     coeffs_cens <- as.numeric(model_cens$coefficients)
     bh_cens <- survival::basehaz(model_cens, centered=F)
@@ -40,7 +43,6 @@ construct_Q_n <- function(type, dat, vals, return_model=F, print_coeffs=F) {
         return(1)
       } else {
         Lambda_t <- bh_srv$hazard[which.min(abs(bh_srv$time-t))]
-        # Lambda_t <- bh_srv$hazard[max(which((bh_srv$time<t)==T))] # This version sometimes throws an error
         return(exp(-1*Lambda_t*exp(sum(coeffs_srv*as.numeric(c(x,s))))))
       }
     }
@@ -50,7 +52,6 @@ construct_Q_n <- function(type, dat, vals, return_model=F, print_coeffs=F) {
         return(1)
       } else {
         Lambda_t <- bh_cens$hazard[which.min(abs(bh_cens$time-t))]
-        # Lambda_t <- bh_cens$hazard[max(which((bh_cens$time<t)==T))] # This version sometimes throws an error
         return(exp(-1*Lambda_t*exp(sum(coeffs_cens*as.numeric(c(x,s))))))
       }
     }
@@ -68,6 +69,13 @@ construct_Q_n <- function(type, dat, vals, return_model=F, print_coeffs=F) {
     methods <- c("survSL.coxph", "survSL.expreg", "survSL.km",
                  "survSL.loglogreg", "survSL.pchreg", "survSL.weibreg")
 
+    # Prevents CRAN Note
+    if (F) {
+      x <- earth::earth
+      x <- glmnet::glmnet
+      x <- e1071::svm
+    }
+
     newX <- cbind(vals$x, s=vals$s)[which(vals$t==0),]
     new.times <- unique(vals$t)
 
@@ -85,79 +93,35 @@ construct_Q_n <- function(type, dat, vals, return_model=F, print_coeffs=F) {
     )
 
     srv <- survSuperLearner(
-      time = dat$y,
-      event = dat$delta,
-      X = cbind(dat$x, s=dat$s),
+      time = dat_v$y,
+      event = dat_v$delta,
+      X = dat_v[,c(1:dim_x,which(names(dat_v)=="s"))],
       newX = newX,
       new.times = new.times,
       event.SL.library = methods,
       cens.SL.library = methods,
-      obsWeights = dat$weights,
+      obsWeights = dat_v$weights,
       control = list(initWeightAlg=methods[1], max.SL.iter=10)
     )
 
-    if (print_coeffs) {
-      cat("\n------------------------------\n")
-      cat("SuperLearner algorithm weights\n")
-      cat("------------------------------\n\n")
-      cat("event.coef\n")
-      cat("----------\n")
-      cat(sort(srv$event.coef, decreasing=T))
-      cat("\ncens.coef\n")
-      cat("---------\n")
-      cat(sort(srv$cens.coef, decreasing=T))
-      cat("\n------------------------------\n")
-    }
-
     srv_pred <- srv$event.SL.predict
     cens_pred <- srv$cens.SL.predict
     rm(srv)
 
-    # !!!!! Later consolidate these via a wrapper/constructor function
     fnc_srv <- function(t, x, s) {
-      r <- list()
-      for (i in 1:length(x)) {
-        r[[i]] <- which(abs(x[i]-newX[[paste0("x",i)]])<1e-8)
-      }
-      if (methods::is(newX[["s"]][1],"factor")) {
-        r[[length(x)+1]] <- which(s==newX[["s"]])
-      } else {
-        r[[length(x)+1]] <- which(abs(s-newX[["s"]])<1e-8)
-      }
-      row <- Reduce(intersect, r)
+      row <- find_index(c(x,s), newX)
       col <- which.min(abs(t-new.times))
-      if (length(row)!=1) {
-        stop(paste0("Error in Q_n (B); ", "t=",t,",x=(",
-                    paste(x,collapse=","),"),s=",s,""))
-      }
-      if (length(col)!=1) {
-        stop(paste0("Error in Q_n (C); ", "t=",t,",x=(",
-                    paste(x,collapse=","),"),s=",s,""))
-      }
+      if (length(col)!=1) { stop("length(col)!=1") }
       return(srv_pred[row,col])
     }
 
     fnc_cens <- function(t, x, s) {
-      r <- list()
-      for (i in 1:length(x)) {
-        r[[i]] <- which(abs(x[i]-newX[[paste0("x",i)]])<1e-8)
-      }
-      if (methods::is(newX[["s"]][1],"factor")) {
-        r[[length(x)+1]] <- which(s==newX[["s"]])
-      } else {
-        r[[length(x)+1]] <- which(abs(s-newX[["s"]])<1e-8)
-      }
-      row <- Reduce(intersect, r)
+      row <- find_index(c(x,s), newX)
       col <- which.min(abs(t-new.times))
-      if (length(row)!=1) {
-        stop(paste0("Error in Q_n (B); ", "t=",t,",x=(",
-                    paste(x,collapse=","),"),s=",s,""))
-      }
-      if (length(col)!=1) {
-        stop(paste0("Error in Q_n (C); ", "t=",t,",x=(",
-                    paste(x,collapse=","),"),s=",s,""))
-      }
-      return(cens_pred[row,col])
+      if (length(col)!=1) { stop("length(col)!=1") }
+      # Note: the max() function is to prevent unreasonably small censoring
+      #       probabilities from destabilizing estimates
+      return(max(cens_pred[row,col], 0.001))
     }
 
   }
@@ -168,9 +132,9 @@ construct_Q_n <- function(type, dat, vals, return_model=F, print_coeffs=F) {
     new.times <- unique(vals$t)
 
     survML_args <- list(
-      time = dat$y,
-      event = dat$delta,
-      X = cbind(dat$x, s=dat$s),
+      time = dat_v$y,
+      event = dat_v$delta,
+      X = dat_v[,c(1:dim_x,which(names(dat_v)=="s"))],
       newX = newX,
       newtimes = new.times,
       bin_size = 0.05,
@@ -179,85 +143,171 @@ construct_Q_n <- function(type, dat, vals, return_model=F, print_coeffs=F) {
         # SL.library = rep(c("SL.mean", "SL.glm", "SL.gam", "SL.earth"),2), # Note: rep() is to avoid a SuperLearner bug
         SL.library = rep(c("SL.mean", "SL.glm", "SL.gam"),2), # Note: rep() is to avoid a SuperLearner bug
         V = 5,
-        obsWeights = dat$weights
+        obsWeights = dat_v$weights
       )
     )
     if (type=="survML-G") {
-      # fit <- do.call(survML::stackG, survML_args)
-      fit <- do.call(stackG, survML_args)
+
+      fit <- do.call(survML::stackG, survML_args)
       srv_pred <- fit$S_T_preds
       cens_pred <- fit$S_C_preds
+
     } else if (type=="survML-L") {
+
       survML_args2 <- survML_args
       survML_args2$event <- round(1 - survML_args2$event)
-      # fit_s <- do.call(survML::stackL, survML_args)
-      fit_s <- do.call(stackL, survML_args)
-      # fit_c <- do.call(survML::stackL, survML_args2)
-      fit_c <- do.call(stackL, survML_args2)
+      fit_s <- do.call(survML::stackL, survML_args)
+      fit_c <- do.call(survML::stackL, survML_args2)
       srv_pred <- fit_s$S_T_preds
       cens_pred <- fit_c$S_T_preds
-    }
 
-    if (print_coeffs) {
-      cat("\n-------------------------------------\n")
-      cat("survML SuperLearner algorithm weights\n")
-      cat("\n-------------------------------------\n")
-      cat("P_Delta coeffs\n")
-      cat("--------------\n")
-      fit$fits$P_Delta$reg.object$coef
-      cat("S_Y_1 coeffs\n")
-      cat("------------\n")
-      fit$fits$S_Y_1$reg.object$coef
-      cat("S_Y_0 coeffs\n")
-      cat("------------\n")
-      fit$fits$S_Y_0$reg.object$coef
-      cat("\n-------------------------------------\n")
     }
 
     fnc_srv <- function(t, x, s) {
-      r <- list()
-      for (i in 1:length(x)) {
-        r[[i]] <- which(abs(x[i]-newX[[paste0("x",i)]])<1e-8)
-      }
-      if (methods::is(newX[["s"]][1],"factor")) {
-        r[[length(x)+1]] <- which(s==newX[["s"]])
-      } else {
-        r[[length(x)+1]] <- which(abs(s-newX[["s"]])<1e-8)
-      }
-      row <- Reduce(intersect, r)
+      row <- find_index(c(x,s), newX)
       col <- which.min(abs(t-new.times))
-      if (length(row)!=1) {
-        stop(paste0("Error in Q_n (B); ", "t=",t,",x=(",
-                    paste(x,collapse=","),"),s=",s,""))
-      }
-      if (length(col)!=1) {
-        stop(paste0("Error in Q_n (C); ", "t=",t,",x=(",
-                    paste(x,collapse=","),"),s=",s,""))
-      }
+      if (length(col)!=1) { stop("length(col)!=1") }
       return(srv_pred[row,col])
     }
 
     fnc_cens <- function(t, x, s) {
-      r <- list()
-      for (i in 1:length(x)) {
-        r[[i]] <- which(abs(x[i]-newX[[paste0("x",i)]])<1e-8)
-      }
-      if (methods::is(newX[["s"]][1],"factor")) {
-        r[[length(x)+1]] <- which(s==newX[["s"]])
-      } else {
-        r[[length(x)+1]] <- which(abs(s-newX[["s"]])<1e-8)
-      }
-      row <- Reduce(intersect, r)
+      row <- find_index(c(x,s), newX)
       col <- which.min(abs(t-new.times))
-      if (length(row)!=1) {
-        stop(paste0("Error in Q_n (B); ", "t=",t,",x=(",
-                    paste(x,collapse=","),"),s=",s,""))
+      if (length(col)!=1) { stop("length(col)!=1") }
+      # Note: the max() function is to prevent unreasonably small censoring
+      #       probabilities from destabilizing estimates
+      return(max(cens_pred[row,col], 0.001))
+    }
+
+  }
+
+  if (type=="Cox") {
+    sfnc_srv <- fnc_srv
+    sfnc_cens <- fnc_cens
+  } else {
+    sfnc_srv <- memoise2(fnc_srv)
+    sfnc_cens <- memoise2(fnc_cens)
+  }
+  rm("vals", envir=environment(get("fnc_srv",envir=environment(sfnc_srv))))
+
+  return(list(srv=sfnc_srv, cens=sfnc_cens))
+
+}
+
+
+
+#' Construct conditional survival estimator Q_n
+#'
+#' @param type One of c("true", "Cox", "Random Forest", "Super Learner")
+#' @param dat Subsample of dataset returned by `ss` for which z==1
+#' @param vals List of values to pre-compute function on; REQUIRED FOR SUPERLEARNER
+#' @param return_model Logical; if TRUE, return the model object instead of the
+#'     function
+#' @return Conditional density estimator function
+#'
+#' @noRd
+construct_Q_noS_n <- function(type, dat, vals, return_model=F) {
+
+  dim_x <- attr(dat,"dim_x")
+
+  # if (type=="Cox") {
+  if (type %in% c("Cox", "survML-G", "survML-L")) { # !!!!! Temporary to avoid R-CMD-CHECK failure
+
+    x_names <- names(dat)[1:dim_x]
+    dat$delta2 <- 1 - dat$delta
+
+    model_srv <- survival::coxph(
+      formula = stats::formula(paste0("survival::Surv(y,delta)~",
+                                      paste(x_names,collapse="+"))),
+      data = dat
+    )
+    coeffs_srv <- as.numeric(model_srv$coefficients)
+    bh_srv <- survival::basehaz(model_srv, centered=F)
+
+    model_cens <- survival::coxph(
+      formula = stats::formula(paste0("survival::Surv(y,delta2)~",
+                                      paste(x_names,collapse="+"))),
+      data = dat
+    )
+    coeffs_cens <- as.numeric(model_cens$coefficients)
+    bh_cens <- survival::basehaz(model_cens, centered=F)
+
+    fnc_srv <- function(t, x) {
+      if (t==0) {
+        return(1)
+      } else {
+        Lambda_t <- bh_srv$hazard[which.min(abs(bh_srv$time-t))]
+        return(exp(-1*Lambda_t*exp(sum(coeffs_srv*as.numeric(x)))))
       }
-      if (length(col)!=1) {
-        stop(paste0("Error in Q_n (C); ", "t=",t,",x=(",
-                    paste(x,collapse=","),"),s=",s,""))
+    }
+
+    fnc_cens <- function(t, x) {
+      if (t==0) {
+        return(1)
+      } else {
+        Lambda_t <- bh_cens$hazard[which.min(abs(bh_cens$time-t))]
+        return(exp(-1*Lambda_t*exp(sum(coeffs_cens*as.numeric(x)))))
       }
-      return(cens_pred[row,col])
+    }
+
+    rm(model_srv)
+    rm(model_cens)
+
+  } else {
+    do.call("library", list("SuperLearner"))
+  }
+
+  if (type=="survSL") {
+
+    # Excluding "survSL.rfsrc" for now. survSL.pchSL gives errors.
+    methods <- c("survSL.coxph", "survSL.expreg", "survSL.km",
+                 "survSL.loglogreg", "survSL.pchreg", "survSL.weibreg")
+
+    newX <- vals$x[which(vals$t==0),, drop=F]
+    new.times <- unique(vals$t)
+
+    # Temporary (until survSuperLearner is on CRAN)
+    survSuperLearner <- function() {}
+    rm(survSuperLearner)
+    tryCatch(
+      expr = { do.call("library", list("survSuperLearner")) },
+      error = function(e) {
+        stop(paste0(
+          "To use surv_type='survSL', you must install the `survSuperLearner` ",
+          "package from github, using:\n\ndevtools::install_github(repo='tedwe",
+          "stling/survSuperLearner')"))
+      }
+    )
+
+    srv <- survSuperLearner(
+      time = dat$y,
+      event = dat$delta,
+      X = dat[,c(1:dim_x), drop=F],
+      newX = newX,
+      new.times = new.times,
+      event.SL.library = methods,
+      cens.SL.library = methods,
+      control = list(initWeightAlg=methods[1], max.SL.iter=10)
+    )
+
+    srv_pred <- srv$event.SL.predict
+    cens_pred <- srv$cens.SL.predict
+    rm(srv)
+
+    fnc_srv <- function(t, x) {
+      row <- find_index(x, newX)
+      col <- which.min(abs(t-new.times))
+      if (length(col)!=1) { stop("length(col)!=1") }
+      return(srv_pred[row,col])
+    }
+
+    fnc_cens <- function(t, x) {
+      row <- find_index(x, newX)
+      col <- which.min(abs(t-new.times))
+      if (length(col)!=1) { stop("length(col)!=1") }
+      # Note: the max() function is to prevent unreasonably small censoring
+      #       probabilities from destabilizing estimates
+      return(max(cens_pred[row,col], 0.001))
     }
 
   }
@@ -310,7 +360,7 @@ construct_omega_n <- function(Q_n, Qc_n, t_0, grid) {
       return(sum(x_vals(x,s)[1:index]*y_vals(x,s)[1:index]))
     }
   }
-  # omega_integral <- memoise2(omega_integral) # !!!!! Maybe try un-memoising this
+  # omega_integral <- memoise2(omega_integral) # !!!!! Does this make a difference (for profiling)?
 
   fnc <- function(x,s,y,delta) {
     Q_n(t_0,x,s) * (
@@ -325,9 +375,59 @@ construct_omega_n <- function(Q_n, Qc_n, t_0, grid) {
 
 
 
+#' Construct estimator of nuisance influence function omega_n
+#'
+#' @param Q_noS_n Conditional survival function estimator returned by
+#'     `construct_Q_noS_n` (no biomarker)
+#' @param Qc_noS_n Conditional censoring survival function estimator returned by
+#'     `construct_Q_noS_n` (no biomarker)
+#' @param type Defaults to "estimated". Override with "true" for debugging. Note
+#'     that type="true" only works for surv_true="Cox" and assumes that Q_0
+#'     and Qc_0 (i.e. the true functions) are passed in.
+#' @return Estimator function of nuisance omega_0
+#' @noRd
+construct_omega_noS_n <- function(Q_noS_n, Qc_noS_n, t_0, grid) {
+
+  Q_n <- Q_noS_n
+  Qc_n <- Qc_noS_n
+
+  # Construct cumulative hazard estimator
+  H_n <- function(t,x) { -1 * log(Q_n(t,x)) }
+
+  # Construct memoised integral function
+  x_vals <- memoise2(function(x) {
+    diff(sapply(grid$y, function(t) { H_n(t,x) }))
+  })
+  y_vals <- memoise2(function(x) {
+    sapply(grid$y[2:length(grid$y)], function(t) {
+      ( Q_n(t,x) * Qc_n(t,x) )^-1
+    })
+  })
+  omega_integral <- function(k,x) {
+    index <- which.min(abs(k-grid$y)) - 1
+    if (index==0) {
+      return(0)
+    } else {
+      return(sum(x_vals(x)[1:index]*y_vals(x)[1:index]))
+    }
+  }
+
+  fnc <- function(x,y,delta) {
+    Q_n(t_0,x) * (
+      (delta * In(y<=t_0)) / (Q_n(y,x) * Qc_n(y,x)) -
+        omega_integral(min(y,t_0),x)
+    )
+  }
+
+  return(memoise2(fnc))
+
+}
+
+
+
 #' Construct estimator of conditional density of S given X
 #'
-#' @param dat Subsample of dataset returned by `ss` for which z==1
+#' @param dat_v2 Dataset returned by `load_data`, subset to ph2 vacc
 #' @param type One of c("parametric", "binning")
 #' @param k Number of bins for the binning estimator (if k=0, then the number of
 #'     bins will be selected via cross-validation); ignored for the parametric
@@ -337,9 +437,12 @@ construct_omega_n <- function(Q_n, Qc_n, t_0, grid) {
 #' @note
 #'   - Assumes support of S is [0,1]
 #' @noRd
-construct_f_sIx_n <- function(dat, type, k=0, z1=F) {
+construct_f_sIx_n <- function(dat_v2, type, k=0, z1=F) {
 
-  if (z1) { dat$weights <- rep(1, length(dat$weights)) }
+  dim_x <- attr(dat_v2, "dim_x")
+  n_vacc2 <- attr(dat_v2, "n_vacc2")
+
+  if (z1) { dat_v2$weights <- rep(1, length(dat_v2$weights)) }
 
   if (type=="parametric") {
 
@@ -352,15 +455,14 @@ construct_f_sIx_n <- function(dat, type, k=0, z1=F) {
 
     # Set up weighted likelihood
     wlik <- function(prm) {
-      -1 * sum(sapply(c(1:length(dat$s)), function(i) {
-        dat$weights[i] *
-          log(pmax(dens_s(s=dat$s[i], x=as.numeric(dat$x[i,]), prm),1e-8))
+      -1 * sum(dat_v2$weights * apply(dat_v2, 1, function(r) {
+        log(pmax(dens_s(s=r[["s"]], x=as.numeric(r[1:dim_x]), prm), 1e-8))
       }))
     }
 
     # Run optimizer
     opt <- Rsolnp::solnp(
-      pars = c(rep(0, length(dat$x)+1), 0.15),
+      pars = c(rep(0, dim_x+1), 0.15),
       fun = wlik
     )
     if (opt$convergence!=0) {
@@ -369,7 +471,7 @@ construct_f_sIx_n <- function(dat, type, k=0, z1=F) {
     prm <- opt$pars
 
     # Remove large intermediate objects
-    rm(dat,dens_s,opt)
+    rm(dat_v2,dens_s,opt)
 
     fnc <- function(s, x) {
       mu <- sum( c(1,x) * c(expit(prm[1]),prm[2:(length(x)+1)]) )
@@ -379,54 +481,6 @@ construct_f_sIx_n <- function(dat, type, k=0, z1=F) {
 
   }
 
-  if (type=="parametric (edge)") {
-
-    # Density for a single observation
-    dens_s <- function(s, x, prm) {
-      mu <- sum( c(1,x) * c(expit(prm[1]),prm[2:(length(x)+1)]) )
-      sigma <- 10*expit(prm[length(prm)])
-      return( truncnorm::dtruncnorm(s, a=0.01, b=1, mean=mu, sd=sigma) )
-    }
-
-    # Estimate p_0
-    n_orig <- attr(dat, "n_orig")
-    p_n <- (1/n_orig) * sum(dat$weights * In(dat$s!=0))
-
-    # Filter out observations with s==0
-    dat <- ss(dat, which(dat$s!=0))
-
-    # Set up weighted likelihood
-    wlik <- function(prm) {
-      -1 * sum(sapply(c(1:length(dat$s)), function(i) {
-        dat$weights[i] *
-          log(pmax(dens_s(s=dat$s[i], x=as.numeric(dat$x[i,]), prm),1e-8))
-      }))
-    }
-
-    # Run optimizer
-    opt <- Rsolnp::solnp(
-      pars = c(rep(0, length(dat$x)+1), 0.15),
-      fun = wlik
-    )
-    if (opt$convergence!=0) {
-      warning("f_sIx_n: Rsolnp::solnp() did not converge")
-    }
-    prm <- opt$pars
-
-    # Remove large intermediate objects
-    rm(dat,dens_s,opt)
-
-    fnc <- function(s, x) {
-      mu <- sum( c(1,x) * c(expit(prm[1]),prm[2:(length(x)+1)]) )
-      sigma <- 10*expit(prm[length(prm)])
-      pc_1 <- In(s==0) * ( (1-p_n) / 0.01 )
-      pc_2 <- In(s!=0) * p_n *
-        truncnorm::dtruncnorm(s, a=0, b=1, mean=mu, sd=sigma)
-      return(pc_1+pc_2)
-    }
-
-  }
-
   if (type=="parametric (edge) 2") {
 
     # Density for a single observation
@@ -445,38 +499,34 @@ construct_f_sIx_n <- function(dat, type, k=0, z1=F) {
 
     # Set up weighted likelihood (edge)
     wlik_1 <- function(prm) {
-      -1 * sum(sapply(c(1:length(dat$s)), function(i) {
-        dat$weights[i] *
-          log(pmax(prob_s(s=dat$s[i], x=as.numeric(dat$x[i,]), prm),1e-8))
+      -1 * sum(dat_v2$weights * apply(dat_v2, 1, function(r) {
+        log(pmax(prob_s(s=r[["s"]], x=as.numeric(r[1:dim_x]), prm), 1e-8))
       }))
     }
 
     # Run optimizer (edge)
     opt_1 <- Rsolnp::solnp(
-      pars = rep(0.01, length(dat$x)+1),
+      pars = rep(0.01, dim_x+1),
       fun = wlik_1
     )
     if (opt_1$convergence!=0) {
       warning("f_sIx_n: Rsolnp::solnp did not converge")
     }
     prm_1 <- opt_1$pars
-    # print("prm_1") # !!!!!
-    # print(prm_1) # !!!!!
 
     # Filter out observations with s==0
-    dat_1 <- ss(dat, which(dat$s!=0))
+    dat_1 <- dat_v2[dat_v2$s!=0,]
 
     # Set up weighted likelihood (Normal)
     wlik_2 <- function(prm) {
-      -1 * sum(sapply(c(1:length(dat_1$s)), function(i) {
-        dat_1$weights[i] *
-          log(pmax(dens_s(s=dat_1$s[i], x=as.numeric(dat_1$x[i,]), prm),1e-8))
+      -1 * sum(dat_1$weights * apply(dat_1, 1, function(r) {
+        log(pmax(dens_s(s=r[["s"]], x=as.numeric(r[c(1:dim_x)]), prm), 1e-8))
       }))
     }
 
     # Run optimizer (Normal)
     opt_2 <- Rsolnp::solnp(
-      pars = c(rep(0.01, length(dat_1$x)+1), 0.15),
+      pars = c(rep(0.01, dim_x+1), 0.15),
       fun = wlik_2
     )
     if (opt_2$convergence!=0) {
@@ -485,7 +535,7 @@ construct_f_sIx_n <- function(dat, type, k=0, z1=F) {
     prm_2 <- opt_2$pars
 
     # Remove large intermediate objects
-    rm(dat,dat_1,opt_1,opt_2)
+    rm(dat_v2,dat_1,opt_1,opt_2)
 
     fnc <- function(s, x) {
       if (s==0) {
@@ -502,7 +552,7 @@ construct_f_sIx_n <- function(dat, type, k=0, z1=F) {
     # Set up binning density (based on Diaz and Van Der Laan 2011)
     # prm[1] through prm[k-1] are the hazard components for the bins 1 to k-1
     # prm[k] and prm[k+1] are the coefficients for W1 and W2
-    create_dens <- function(k, dat, remove_dat=F) {
+    create_dens <- function(k, dat_v2, remove_dat=F) {
 
       # Cut points
       alphas <- seq(0, 1, length.out=k+1)
@@ -520,10 +570,8 @@ construct_f_sIx_n <- function(dat, type, k=0, z1=F) {
       })
 
       # Set up weighted likelihood
-      dat_df <- as_df(dat)
-      dim_x <- attr(dat, "dim_x")
       wlik <- function(prm) {
-        -1 * sum(dat$weights * log(pmax(apply(dat_df, 1, function(r) {
+        -1 * sum(dat_v2$weights * log(pmax(apply(dat_v2, 1, function(r) {
           dens_s(s=r[["s"]], x=as.numeric(r[1:dim_x]), prm)
         }),1e-8)))
       }
@@ -531,7 +579,7 @@ construct_f_sIx_n <- function(dat, type, k=0, z1=F) {
       # Run optimizer
       if(.Platform$OS.type=="unix") { sink("/dev/null") } else { sink("NUL") }
       opt <- Rsolnp::solnp(
-        pars = rep(0.001,k+length(dat$x)-1),
+        pars = rep(0.001,k+dim_x-1),
         fun = wlik
       )
       sink()
@@ -543,7 +591,7 @@ construct_f_sIx_n <- function(dat, type, k=0, z1=F) {
       # Remove large intermediate objects
       rm(dens_s,opt)
 
-      if (remove_dat) { rm(dat) }
+      if (remove_dat) { rm(dat_v2) }
 
       fnc <- function(s, x) {
 
@@ -565,8 +613,8 @@ construct_f_sIx_n <- function(dat, type, k=0, z1=F) {
 
       # Prep
       n_folds <- 5
-      folds <- sample(cut(c(1:length(dat$s)), breaks=n_folds, labels=FALSE))
-      ks <- c(5,10,15,20) # !!!!! Make this an argument
+      folds <- sample(cut(c(1:n_vacc2), breaks=n_folds, labels=FALSE))
+      ks <- c(4,8,12,16) # !!!!! Make this an argument
       best <- list(k=999, max_log_lik=999)
 
       # Cross-validation
@@ -574,21 +622,12 @@ construct_f_sIx_n <- function(dat, type, k=0, z1=F) {
 
         sum_log_lik <- 0
         for (i in c(1:n_folds)) {
-          dat_train <- list(
-            s = dat$s[-which(folds==i)],
-            weights = dat$weights[-which(folds==i)],
-            x = dat$x[-which(folds==i),]
-          )
-          dat_test <- list(
-            s = dat$s[which(folds==i)],
-            weights = dat$weights[which(folds==i)],
-            x = dat$x[which(folds==i),]
-          )
+          dat_train <- dat_v2[which(folds!=i),]
+          dat_test <- dat_v2[which(folds==i),]
           dens <- create_dens(k, dat_train)
-          sum_log_lik <- sum_log_lik +
-            sum(log(sapply(c(1:length(dat_test$s)), function(j) {
-              dens(dat_test$s[j], as.numeric(dat_test$x[j,]))
-            })))
+          sum_log_lik <- sum_log_lik + sum(log(apply(dat_test, 1, function(r) {
+            dens(r[["s"]], as.numeric(r[c(1:dim_x)]))
+          })))
         }
 
         if (sum_log_lik>best$max_log_lik || best$max_log_lik==999) {
@@ -603,7 +642,7 @@ construct_f_sIx_n <- function(dat, type, k=0, z1=F) {
 
     }
 
-    fnc <- create_dens(k, dat, remove_dat=T)
+    fnc <- create_dens(k, dat_v2, remove_dat=T)
 
   }
 
@@ -620,14 +659,27 @@ construct_f_sIx_n <- function(dat, type, k=0, z1=F) {
 #'     `construct_f_sIx_n`
 #' @return Marginal density estimator function
 #' @noRd
-construct_f_s_n <- function(dat_orig, f_sIx_n) {
+construct_f_s_n <- function(dat_v, f_sIx_n) {
 
-  n_orig <- attr(dat_orig, "n_orig")
-  memoise2(function(s) {
-    (1/n_orig) * sum(apply(dat_orig$x, 1, function(x) {
-      f_sIx_n(s,as.numeric(x))
-    }))
-  })
+  n_vacc <- attr(dat_v, "n_vacc")
+  dim_x <- attr(dat_v, "dim_x")
+
+  if (attr(dat_v, "covariates_ph2")) {
+    dat_v2 <- dat_v[dat_v$z==1,]
+    datx_v2 <- dat_v2[, c(1:dim_x), drop=F]
+    memoise2(function(s) {
+      (1/n_vacc) * sum(dat_v2$weights * apply(datx_v2, 1, function(x) {
+        f_sIx_n(s,as.numeric(x))
+      }))
+    })
+  } else {
+    datx_v <- dat_v[, c(1:dim_x), drop=F]
+    memoise2(function(s) {
+      (1/n_vacc) * sum(apply(datx_v, 1, function(x) {
+        f_sIx_n(s,as.numeric(x))
+      }))
+    })
+  }
 
 }
 
@@ -654,61 +706,17 @@ construct_g_n <- function(f_sIx_n, f_s_n) {
 #' @param type One of c("step", "linear (mid)")
 #' @return IPS-weighted CDF estimator function
 #' @noRd
-construct_Phi_n <- function (dat_orig, dat, type="linear (mid)") {
+construct_Phi_n <- function (dat_v2, dat, type="linear (mid)") {
 
   # !!!!! type currently ignored
 
-  n_orig <- attr(dat_orig, "n_orig")
+  n_vacc <- attr(dat_v2, "n_vacc")
 
-  fn <- memoise::memoise(function(x) { # !!!!! Switch to memoise2
-    (1/n_orig) * sum(dat$weights*In(dat$s<=x))
+  fnc <- memoise2(function(x) {
+    (1/n_vacc) * sum(dat_v2$weights*In(dat_v2$s<=x))
   })
 
-  return(fn)
-
-  # # n_orig <- attr(dat, "n_orig")
-  # n_orig <- sum(dat$weights) # !!!!!
-  # df <- data.frame(s=dat$s, weights=dat$weights)
-  # df <- dplyr::arrange(df, s)
-  # vals_x <- unique(df$s)
-  # vals_y <- c()
-  #
-  # for (j in 1:length(vals_x)) {
-  #   indices <- which(df$s==vals_x[j])
-  #   weights_j <- df$weights[indices]
-  #   new_y_val <- (1/n_orig) * sum(weights_j)
-  #   vals_y <- c(vals_y, new_y_val)
-  # }
-  # vals_y <- cumsum(vals_y)
-  #
-  # if (type=="step") {
-  #   method <- "constant"
-  # } else if (type=="linear (mid)") {
-  #   vals_x <- c(vals_x[1], vals_x[1:(length(vals_x)-1)]+diff(vals_x)/2,
-  #               vals_x[length(vals_x)])
-  #   vals_y <- c(0, vals_y[1:(length(vals_y)-1)], 1)
-  #   method <- "linear"
-  # }
-  #
-  # return(stats::approxfun(vals_x, vals_y, method=method, yleft=0, yright=1, f=0,
-  #                  ties="ordered"))
-
-}
-
-
-
-#' Construct nuisance estimator eta_n
-#'
-#' @noRd
-construct_eta_n <- function(dat, Q_n, t_0) {
-
-  n_orig <- attr(dat, "n_orig")
-
-  return(memoise2(function(u,x) {
-    (1/n_orig) * sum(
-      dat$weights * sapply(dat$s, function(s) { In(s<=u) * (1-Q_n(t_0,x,s)) })
-    )
-  }))
+  return(memoise2(fnc))
 
 }
 
@@ -721,28 +729,26 @@ construct_eta_n <- function(dat, Q_n, t_0) {
 #'     `construct_Q_n`
 #' @return G-computation estimator of theta_0
 #' @noRd
-construct_r_tilde_Mn <- function(dat_orig, Q_n, t_0) {
+construct_r_tilde_Mn <- function(dat_v, Q_n, t_0) {
 
-  n_orig <- attr(dat_orig, "n_orig")
-  memoise2(function(s) {
-    1 - (1/n_orig) * sum(apply(dat_orig$x, 1, function(x) {
-      Q_n(t_0, as.numeric(x), s)
-    }))
-  })
-
-}
-
-
-
-#' Construct nuisance estimator Gamma_tilde_n
-#'
-#' @noRd
-construct_Gamma_tilde_n <- function(dat, r_tilde_Mn) {
-
-  n_orig <- attr(dat, "n_orig")
-  piece_1 <- dat$weights * sapply(dat$s, r_tilde_Mn)
-
-  return(memoise2(function(u) { (1/n_orig) * sum(In(dat$s<=u)*piece_1) }))
+  n_vacc <- attr(dat_v, "n_vacc")
+  dim_x <- attr(dat_v, "dim_x")
+  if (attr(dat_v, "covariates_ph2")) {
+    dat_v2 <- dat_v[dat_v$z==1,]
+    datx_v2 <- dat_v2[, c(1:dim_x), drop=F]
+    memoise2(function(s) {
+      1 - (1/n_vacc) * sum(dat_v2$weights * apply(datx_v2, 1, function(x) {
+        Q_n(t_0, as.numeric(x), s)
+      }))
+    })
+  } else {
+    datx_v <- dat_v[, c(1:dim_x), drop=F]
+    memoise2(function(s) {
+      1 - (1/n_vacc) * sum(apply(datx_v, 1, function(x) {
+        Q_n(t_0, as.numeric(x), s)
+      }))
+    })
+  }
 
 }
 
@@ -754,7 +760,6 @@ construct_Gamma_tilde_n <- function(dat, r_tilde_Mn) {
 construct_f_n_srv <- function(Q_n, Qc_n, grid) {
 
   # Helper function to calculate derivatives
-  # !!!!! Move outside?
   construct_surv_deriv <- function(Q) {
 
     max_index <- length(grid$y)
@@ -799,7 +804,7 @@ construct_f_n_srv <- function(Q_n, Qc_n, grid) {
 #' Construct q_n nuisance estimator function
 #'
 #' @noRd
-construct_q_n <- function(type="standard", dat, omega_n, g_n, r_tilde_Mn,
+construct_q_n <- function(type="standard", dat_v2, omega_n, g_n, r_tilde_Mn,
                           f_n_srv) {
 
   if (type=="standard") {
@@ -809,39 +814,39 @@ construct_q_n <- function(type="standard", dat, omega_n, g_n, r_tilde_Mn,
       omega_n(x,s,y,delta)/g_n(s,x)+r_tilde_Mn(s)
     })
 
-    q_n_star <- memoise2(function(y,delta,x,s,u) {
+    # q_n_star <- memoise2(function(y,delta,x,s,u) {
+    q_n_star <- function(y,delta,x,s,u) {
       if (s>u) {
         return(0)
       } else {
         return(q_n_star_inner(x,y,delta,s))
       }
-    })
-
-    n <- length(dat$s)
+    # })
+    }
 
     # Helper functions
     # !!!!! Can these vectors/functions be used elsewhere?
     f_n_srv_s <- memoise2(function(y,delta,x) {
-      sapply(dat$s, function(s) { f_n_srv(y,delta,x,s) })
+      sapply(dat_v2$s, function(s) { f_n_srv(y,delta,x,s) })
     })
     q_n_star_s <- memoise2(function(y,delta,x,u) {
-      sapply(dat$s, function(s) { q_n_star(y,delta,x,s,u) })
+      sapply(dat_v2$s, function(s) { q_n_star(y,delta,x,s,u) })
     })
     g_n_s <- memoise2(function(x) {
-      sapply(dat$s, function(s) { g_n(s,x) })
+      sapply(dat_v2$s, function(s) { g_n(s,x) })
     })
 
-    fnc <- function(x, y, delta, u) {
+    fnc <- function(x,y,delta,u) {
 
       f_n_srv_s <- f_n_srv_s(y,delta,x)
       q_n_star_s <- q_n_star_s(y,delta,x,u)
       g_n_s <- g_n_s(x)
 
-      denom <- sum(dat$weights*f_n_srv_s*g_n_s)
+      denom <- sum(dat_v2$weights*f_n_srv_s*g_n_s)
       if (denom==0) {
         return (0)
       } else {
-        num <- sum(dat$weights*q_n_star_s*f_n_srv_s*g_n_s)
+        num <- sum(dat_v2$weights*q_n_star_s*f_n_srv_s*g_n_s)
         return(num/denom)
       }
 
@@ -864,18 +869,15 @@ construct_q_n <- function(type="standard", dat, omega_n, g_n, r_tilde_Mn,
 #' Construct estimator of nuisance g_sn
 #'
 #' @noRd
-construct_g_sn <- function(dat, f_n_srv, g_n, p_n) {
+construct_g_sn <- function(dat_v2, f_n_srv, g_n, p_n) {
 
-  n_orig <- attr(dat, "n_orig")
-  dat_df <- as_df(dat)
+  n_vacc <- attr(dat_v2, "n_vacc")
 
   return(memoise2(function(x, y, delta) {
     num <- f_n_srv(y, delta, x, 0) * g_n(s=0,x) * (1-p_n)
-    den <- (1/n_orig) * sum(
-      dat$weights * apply(dat_df, 1, function(r) {
-        f_n_srv(y, delta, x, r[["s"]]) * g_n(r[["s"]],x) # !!!!! Can this be replaced with sapply ?????
-      })
-    )
+    den <- (1/n_vacc) * sum(dat_v2$weights * sapply(dat_v2$s, function(s) {
+      f_n_srv(y,delta,x,s) * g_n(s,x)
+    }))
     if (den==0) { return(0) } else { return(num/den) }
   }))
 
@@ -896,27 +898,35 @@ construct_g_sn <- function(dat, f_n_srv, g_n, p_n) {
 #'     `construct_f_sIx_n` among the observations for which z==1
 #' @return gamma_n nuisance estimator function
 #' @noRd
-construct_gamma_n <- function(dat_orig, dat, type="Super Learner", omega_n,
+construct_gamma_n <- function(dat_v, type="Super Learner", omega_n,
                               grid) {
 
+  dim_x <- attr(dat_v, "dim_x")
+  dat_v2 <- dat_v[dat_v$z==1,]
+  datx_v <- dat_v[, c(1:dim_x), drop=F]
+  class(datx_v) <- "data.frame"
+
   # Construct pseudo-outcomes
-  dat_df <- as_df(dat)
-  dim_x <- attr(dat_orig, "dim_x")
-  omega_n_d <- apply(dat_df, 1, function(r) {
+  omega_n_d <- apply(dat_v2, 1, function(r) {
     omega_n(as.numeric(r[1:dim_x]), r[["s"]], r[["y"]], r[["delta"]])
   })
-  dat_df$po <- (dat$weights*as.numeric(omega_n_d))^2
+  dat_v2$po <- (dat_v2$weights*as.numeric(omega_n_d))^2
 
   # Filter out infinite values
-  if (sum(!is.finite(dat_df$po))!=0) {
-    # dat_df <- dplyr::filter(dat_df, is.finite(po))
-    dat_df <- dat_df[which(is.finite(dat_df$po)),] # !!!!! New code (to avoid R CMD CHECK error)
-    warning(paste("gamma_n:", sum(!is.finite(dat_df$po)),
+  if (sum(!is.finite(dat_v2$po))!=0) {
+    dat_v2 <- dat_v2[which(is.finite(dat_v2$po)),]
+    warning(paste("gamma_n:", sum(!is.finite(dat_v2$po)),
                   "non-finite pseudo-outcome values"))
   }
 
   # Setup
-  x_distinct <- dplyr::distinct(dat_orig$x)
+  if (attr(dat_v, "covariates_ph2")) {
+    datx_v2 <- dat_v2[, c(1:dim_x), drop=F]
+    class(datx_v2) <- "data.frame"
+    x_distinct <- dplyr::distinct(datx_v2)
+  } else {
+    x_distinct <- dplyr::distinct(datx_v)
+  }
   x_distinct <- cbind("x_index"=c(1:nrow(x_distinct)), x_distinct)
   newX <- expand.grid(x_index=x_distinct$x_index, s=grid$s)
   newX <- dplyr::inner_join(x_distinct, newX, by="x_index")
@@ -926,12 +936,13 @@ construct_gamma_n <- function(dat_orig, dat, type="Super Learner", omega_n,
   if (type=="Super Learner") {
 
     # Fit SuperLearner regression
+    do.call("library", list("SuperLearner"))
     SL.library <- c("SL.mean", "SL.gam", "SL.ranger", "SL.earth", "SL.loess",
                     "SL.nnet", "SL.ksvm", "SL.rpartPrune", "SL.svm")
 
     model_sl <- SuperLearner::SuperLearner(
-      Y = dat_df$po,
-      X = cbind(dat_df[,c(1:dim_x), drop=F],s=dat_df$s),
+      Y = dat_v2$po,
+      X = dat_v2[,c(1:dim_x,which(names(dat_v2)=="s"))],
       newX = newX,
       family = "gaussian",
       SL.library = SL.library,
@@ -941,28 +952,14 @@ construct_gamma_n <- function(dat_orig, dat, type="Super Learner", omega_n,
     if (sum(pred<0)!=0) {
       warning(paste("gamma_n:", sum(pred<0), "negative predicted values"))
     }
-    # print(coef(model_sl))
-    rm(model_sl)
 
-    newX$index <- c(1:nrow(newX))
-    reg <- function(x,s) {
-      # Dynamically filter to select index
-      cond <- paste0("round(s,5)==",round(s,5))
-      for (i in c(1:length(x))) {
-        cond <- paste0(cond," & round(x",i,",5)==",round(x[i],5))
-      }
-      index <- (dplyr::filter(newX, eval(parse(text=cond))))$index
-      if (length(index)!=1) {
-        stop(paste0("Error in gamma_n; ", "x=(", paste(x,collapse=","), "), s=",
-                    s))
-      }
-      return(pred[index])
-    }
+    # Construct regression function
+    reg <- function(x,s) { pred[find_index(c(x,s), newX)] }
 
   }
 
   # Remove large intermediate objects
-  rm(dat_orig,dat,omega_n)
+  rm(dat_v,dat_v2,datx_v,omega_n,model_sl)
 
   return(memoise2(reg))
 
@@ -973,11 +970,12 @@ construct_gamma_n <- function(dat_orig, dat, type="Super Learner", omega_n,
 #' Construct estimator of g_z0(x,s) = P(Z=1|X=x,S=s)
 #'
 #' @noRd
-construct_g_zn <- function(dat_orig, type="Super Learner", f_sIx_n,
+construct_g_zn <- function(dat_v, type="Super Learner", f_sIx_n,
                            f_sIx_z1_n) {
 
   # Set library
   if (type=="Super Learner") {
+    do.call("library", list("SuperLearner"))
     SL.library <- c("SL.mean", "SL.gam", "SL.ranger", "SL.earth", "SL.nnet",
                     "SL.glmnet")
   } else if (type=="logistic") {
@@ -985,36 +983,43 @@ construct_g_zn <- function(dat_orig, type="Super Learner", f_sIx_n,
   }
 
   # Create data objects
-  newX <- dplyr::distinct(dat_orig$x)
+  dim_x <- attr(dat_v, "dim_x")
+  dat_v2 <- dat_v[dat_v$z==1,]
+  datx_v <- dat_v[, c(1:dim_x), drop=F]
+  datx_v2 <- dat_v2[, c(1:dim_x), drop=F]
+  class(datx_v) <- "data.frame"
+  class(datx_v2) <- "data.frame"
 
   # Fit SuperLearner regression
-  model_sl <- SuperLearner::SuperLearner(
-    Y = dat_orig$z,
-    X = dat_orig$x,
-    newX = newX,
-    family = "binomial",
-    SL.library = SL.library,
-    verbose = F
-  )
+  if (attr(dat_v, "covariates_ph2")) {
+    newX <- dplyr::distinct(datx_v2)
+    model_sl <- SuperLearner::SuperLearner(
+      Y = dat_v2$z,
+      X = datx_v2,
+      newX = newX,
+      family = "binomial",
+      SL.library = SL.library,
+      verbose = F
+    )
+  } else {
+    newX <- dplyr::distinct(datx_v)
+    model_sl <- SuperLearner::SuperLearner(
+      Y = dat_v$z,
+      X = datx_v,
+      newX = newX,
+      family = "binomial",
+      SL.library = SL.library,
+      verbose = F
+    )
+  }
 
   pred <- as.numeric(model_sl$SL.predict)
-  rm(model_sl)
 
-  # Construct function
-  newX$index <- c(1:nrow(newX))
-  reg <- function(x) {
-    # Dynamically filter to select index
-    cond <- "1==1"
-    for (i in c(1:length(x))) {
-      cond <- paste0(cond," & round(x",i,",5)==",round(x[i],5))
-    }
-    index <- (dplyr::filter(newX, eval(parse(text=cond))))$index
-    if (length(index)!=1) {
-      stop(paste0("Error in g_zn; ",
-                  "x=(", paste(x,collapse=","), ")"))
-    }
-    return(pred[index])
-  }
+  # Construct regression function
+  reg <- function(x) { pred[find_index(x, newX)] }
+
+  # Remove large intermediate objects
+  rm(dat_v,datx_v,model_sl)
 
   return(memoise2(function(x,s) { (f_sIx_z1_n(s,x)/f_sIx_n(s,x))*reg(x) }))
 
@@ -1098,19 +1103,23 @@ construct_deriv_r_Mn <- function(type="m-spline", r_Mn, dir, grid) {
 #' @param f_s_n Density estimator returned by `construct_f_s_n`
 #' @return Chernoff scale factor estimator function
 #' @noRd
-construct_tau_n <- function(deriv_r_Mn, gamma_n, f_sIx_n, g_zn,
-                            dat_orig) {
+construct_tau_n <- function(dat_v, deriv_r_Mn, gamma_n, f_sIx_n, g_zn) {
 
-  n_orig <- attr(dat_orig, "n_orig")
-  x <- dat_orig$x
-  return(Vectorize(function(u) {
-    abs(4*deriv_r_Mn(u) * (1/n_orig) * sum(apply(dat_orig$x, 1, function(x) {
-      x <- as.numeric(x)
+  n_vacc <- attr(dat_v, "n_vacc")
+  dim_x <- attr(dat_v, "dim_x")
+  dat_v2 <- dat_v[dat_v$z==1,]
+
+  if (attr(dat_v, "covariates_ph2")) {
+    dat_apply <- dat_v2
+  } else {
+    dat_apply <- dat_v
+  }
+
+  return(function(u) {
+    abs(4*deriv_r_Mn(u) * (1/n_vacc) * sum(apply(dat_apply, 1, function(r) {
+      x <- as.numeric(r[1:dim_x])
       return((gamma_n(x,u)*g_zn(x,u))/f_sIx_n(u,x))
     })))^(1/3)
-  }))
+  })
 
 }
-
-
-
